/*	Agent.cpp
 *
 *	Robert Rose June 17, 2005
 *
 *	Copyright (c) 2005 Robert Rose
 *
 *	Permission is granted to do whatever you want with this code as long
 *	as the author is given due credit.
 *	
 */

#include "Nogredex.h"
#include <Ogre.h>

#include "Obstacle.h"
#include "Agent.h"

#include "LDebug.h"

//	Protected constructor...
//	
CAgent::CAgent()
{
	_nogre_env = NULL;
	_obj = NULL;

	_thinktimer = 0.0f;
}

CAgent::CAgent(NogredexDemoFrameListener *nogre_env, AgentType type)
{
	_nogre_env = nogre_env;

	Ogre::String matName = "Material/earth";

	if(type == AgentTypeEnemy)
		matName = "Material/basicRed";
	else if(type == AgentTypeObstacle)
		matName = "Material/basicBlue";

	_obj = _nogre_env->addBall(matName, MAT_RUBBER, Vector3(0.0, 2.0, 0.0), 1.0);

	_thinktimer = 0.0f;
	_force = Vector3::ZERO;
}

CAgent::~CAgent(void)
{
	// delete _obj is handled by Nogredex
}

void CAgent::spin(const Vector3 &dir, float force)
{
	Vector3 up(0.0, 1.0, 0.0);

	Vector3 tau = -dir.crossProduct(up);

	tau *= force;

	_obj->actor->addTorque(toNx(tau));

	// fudge factor.. push the agent a tiny bit..
	// gives a nice mario bros "i didnt mean to jump off that cliff" effect
	//push(dir, 1000.0);
}

void CAgent::push(const Vector3 &dir, float force)
{
	Vector3 f = dir * force;
	_obj->actor->addForce(toNx(f));
}


Vector3 CAgent::attract(const CAgent *agent, float threshold, float force)
{
	if(agent == this)
		return Vector3(0.0, 0.0, 0.0);

	Vector3 fellowdirection = agent->getDistance(this);
	float fellowdist = fellowdirection.length();

	if(fellowdist < threshold)
	{
		return fellowdirection * (1.0 / fellowdist) * force;
	}
	
	return Vector3(0.0, 0.0, 0.0);
}


//	Naive A* implemention.  Moves only in one of 8 directions and is only aware
//	of one previous move at a time, so it's possible for it to get confused and
//  backtrack in a circle.  Works OK for certain situations and is pretty fast.
//
bool CAgent::pathfind_naive(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g, int prev)
{
	// are we already there?
	Vector3 playerdist = player.getPosition() - pos;
	if(playerdist.length() < 2.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched too far?
	if(g > 40.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	vector<AgentPathSegment> possibilities;

	possibilities.push_back(AgentPathSegment(Vector3(1,0,0)));
	possibilities.push_back(AgentPathSegment(Vector3(1,0,1)));
	possibilities.push_back(AgentPathSegment(Vector3(0,0,1)));
	possibilities.push_back(AgentPathSegment(Vector3(-1,0,1)));
	possibilities.push_back(AgentPathSegment(Vector3(-1,0,0)));
	possibilities.push_back(AgentPathSegment(Vector3(-1,0,-1)));
	possibilities.push_back(AgentPathSegment(Vector3(0,0,-1)));
	possibilities.push_back(AgentPathSegment(Vector3(1,0,-1)));

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		// store the number of this solution so we don't ever go backwards
		possibilities[i].i = i;

		// determine where we would be if we took this path
		Vector3 newpos = pos + possibilities[i].path;


		bool badpossibility = false;

		// check to see if this path would step backwards
		if(possibilities[i].i == prev)
			badpossibility = true;

		// check to see if this path would collide with an obstacle
		Ray ray(pos, possibilities[i].path);

		for(unsigned int j = 0; j < obstacles.size(); j++)
		{
			pair< bool, Real > intersection = ray.intersects(obstacles[j]->getBox());
			
			if(intersection.first && intersection.second < 1.001)
				badpossibility = true;
		}

		// calculate A* parameters
		float pg = g + possibilities[i].path.length();
		float ph = (newpos - player.getPosition()).length();

		float pf = pg + ph;

		if(badpossibility)
			pf = 100000.0;

		possibilities[i].f = pf;
		possibilities[i].g = pg;
	}

	sort(possibilities.begin(), possibilities.end());

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		bool solnfound = pathfind_naive(pos + possibilities[i].path, player, obstacles, visForces, force, possibilities[i].g, (possibilities[i].i + 4) % 8);

		if(solnfound)
		{
			force = possibilities[i].path;

			pair<Vector3, Vector3> vis(pos, pos + force);
			visForces.push_back(vis);

			return true;
		}
	}

	return false;
}


//	Naive A* implemention.  Moves only in one of 4 directions and is only aware
//	of one previous move at a time, so it's possible for it to get confused and
//  backtrack in a circle.  Works OK for certain situations and is pretty fast.
//
bool CAgent::pathfind_naive4(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, vector<AgentSearchPoint> &prev, float g)
{
	// are we already there?
	Vector3 playerdist = player.getPosition() - pos;
	if(playerdist.length() < 2.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched too far?
	if(g > 80.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	vector<AgentPathSegment> possibilities;

	possibilities.push_back(AgentPathSegment(Vector3(1,0,0)));
	possibilities.push_back(AgentPathSegment(Vector3(0,0,1)));
	possibilities.push_back(AgentPathSegment(Vector3(-1,0,0)));
	possibilities.push_back(AgentPathSegment(Vector3(0,0,-1)));

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		// determine where we would be if we took this path
		Vector3 newpos = pos + possibilities[i].path;


		bool badpossibility = false;

		// check to see if this path would step backwards
		AgentSearchPoint p;
		p.x = newpos.x;
		p.z = newpos.z;

		for(unsigned int s = 0; s < prev.size(); s++)
			if(p == prev[s])
				badpossibility = true;

		// check to see if this path would collide with an obstacle
		Ray ray(pos, possibilities[i].path);

		for(unsigned int j = 0; j < obstacles.size(); j++)
		{
			pair< bool, Real > intersection = ray.intersects(obstacles[j]->getBox());
			
			if(intersection.first && intersection.second < 1.001)
				badpossibility = true;
		}

		// calculate A* parameters
		float pg = g + possibilities[i].path.length();
		float ph = (newpos - player.getPosition()).length();

		float pf = pg + ph;

		if(badpossibility)
			pf = 100000.0;

		possibilities[i].f = pf;
		possibilities[i].g = pg;
	}

	sort(possibilities.begin(), possibilities.end());

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		AgentSearchPoint p;
		p.x = pos.x + possibilities[i].path.x;
		p.z = pos.z + possibilities[i].path.z;
		prev.push_back(p);

		bool solnfound = pathfind_naive4(pos + possibilities[i].path, player, obstacles, visForces, force, prev, possibilities[i].g);

		if(solnfound)
		{
			force = possibilities[i].path;

			pair<Vector3, Vector3> vis(pos, pos + force);
			visForces.push_back(vis);

			return true;
		}

		prev.pop_back();
	}

	return false;
}

//	Ray-based A* implementation.  Pretty time consuming; needs muchos optimizationos.
//	Works by casting path rays to the edges of known obstacles and using those points
//	for building paths rather than a grid.  Works well for dense obstacles but is slow.
//	Has issues if two obstacles are adjacent because it's not smart enough to detect a
//	backtrack.
//
bool CAgent::pathfind_ray(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g, int depth)
{
	// are we already there?
	Vector3 playerdist = player.getPosition() - pos;
	if(playerdist.length() < 2.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched too far?
	if(g > 200.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched for too long?
	if(depth > 4)
	{
		force = Vector3(0,0,0);
		return true;
	}

	vector<AgentPathSegment> possibilities;

	// the first possibility is the one that goes straight to the player
	// (even though it might not be possible...)
	possibilities.push_back(AgentPathSegment(playerdist));

	// the other possibilities are the ones that flank all of the obstructions
	// generate possibilities for these flanking scenarios
	for(unsigned int j = 0; j < obstacles.size(); j++)
	{
		if((obstacles[j]->getPosition() - pos).length() < 20.0)
		{
			AxisAlignedBox aab = obstacles[j]->getBox();
			const Vector3 *corners = aab.getAllCorners();

			// iterate through all of the corners of this obstacle
			for(int i = 0; i < 8; i++)
			{
				Vector3 corner = corners[i];

				// naively ignore corners that have a y component (2D approximation)
				if(!(corner.y > 0.0))
				{
					// project out the edges of the corner slightly
					corner += (corner - obstacles[j]->getPosition()).normalisedCopy() * 0.7;

					corner.y = 1.0;
					Vector3 cornerdir = pos - corner;

					// make sure we're not moving to the same corner we're on now
					if(cornerdir.length() > 1.0)
					{
						// calculate the path
						Vector3 path = corner - pos;

						// this could be a possibility...
						// obviously this path could intersect the object it's trying to avoid,
						// but we'll perform that check later
						possibilities.push_back(AgentPathSegment(path));
					}
				}
			}
		}
	}

	// does this path intersect any other obstacles?  that would be bad..
	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		Ray ray(pos, possibilities[i].path);
		bool badpossibility = false;

		for(unsigned int k = 0; k < obstacles.size(); k++)
		{
			pair< bool, Real > intersection = ray.intersects(obstacles[k]->getBox());

			if(intersection.first && intersection.second < 1.1)
			{
				badpossibility = true;
			}
		}

		float pg = g + possibilities[i].path.length();
		float ph = (pos + possibilities[i].path - player.getPosition()).length();

		float pf = pg + ph;

		if(badpossibility)
			pf = 100000.0;

		possibilities[i].f = pf;
		possibilities[i].g = pg;
	}


	sort(possibilities.begin(), possibilities.end());

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		// sanity check.. don't try the obviously bad possibilities
		if(possibilities[i].f < 1000.0)
		{
			bool solnfound = pathfind_ray(pos + possibilities[i].path, player, obstacles, visForces, force, possibilities[i].g, depth+1);

			if(solnfound)
			{
				force = possibilities[i].path;

				pair<Vector3, Vector3> vis(pos, pos + force);
				visForces.push_back(vis);

				return true;
			}
		}
	}

	force = Vector3(0,0,0);
	return false;
}

//	Combination Ray-NaiveGrid A*.  First path is a ray, everything else past that is on the grid.
//
bool CAgent::pathfind_raycombo(const Vector3 &pos, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, Vector3 &force, float g, int depth)
{
	// are we already there?
	Vector3 playerdist = player.getPosition() - pos;
	if(playerdist.length() < 2.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched too far?
	if(g > 200.0)
	{
		force = Vector3(0,0,0);
		return true;
	}

	// have we searched for too long?
	if(depth > 4)
	{
		force = Vector3(0,0,0);
		return true;
	}

	vector<AgentPathSegment> possibilities;

	// the first possibility is the one that goes straight to the player
	// (even though it might not be possible...)
	possibilities.push_back(AgentPathSegment(playerdist));

	// the other possibilities are the ones that flank all of the obstructions
	// generate possibilities for these flanking scenarios
	for(unsigned int j = 0; j < obstacles.size(); j++)
	{
		if((obstacles[j]->getPosition() - pos).length() < 20.0)
		{
			AxisAlignedBox aab = obstacles[j]->getBox();
			const Vector3 *corners = aab.getAllCorners();

			// iterate through all of the corners of this obstacle
			for(int i = 0; i < 8; i++)
			{
				Vector3 corner = corners[i];

				// naively ignore corners that have a y component (2D approximation)
				if(!(corner.y > 0.0))
				{
					// project out the edges of the corner slightly
					corner += (corner - obstacles[j]->getPosition()).normalisedCopy() * 0.7;

					corner.y = 1.0;
					Vector3 cornerdir = pos - corner;

					// make sure we're not moving to the same corner we're on now
					if(cornerdir.length() > 1.0)
					{
						// calculate the path
						Vector3 path = corner - pos;

						// this could be a possibility...
						// obviously this path could intersect the object it's trying to avoid,
						// but we'll perform that check later
						possibilities.push_back(AgentPathSegment(path));
					}
				}
			}
		}
	}

	// does this path intersect any other obstacles?  that would be bad..
	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		Ray ray(pos, possibilities[i].path);
		bool badpossibility = false;

		for(unsigned int k = 0; k < obstacles.size(); k++)
		{
			pair< bool, Real > intersection = ray.intersects(obstacles[k]->getBox());

			if(intersection.first && intersection.second < 1.1)
			{
				badpossibility = true;
			}
		}

		float pg = g + possibilities[i].path.length();
		float ph = (pos + possibilities[i].path - player.getPosition()).length();

		float pf = pg + ph;

		if(badpossibility)
			pf = 100000.0;

		possibilities[i].f = pf;
		possibilities[i].g = pg;
	}


	sort(possibilities.begin(), possibilities.end());

	for(unsigned int i = 0; i < possibilities.size(); i++)
	{
		// sanity check.. don't try the obviously bad possibilities
		if(possibilities[i].f < 1000.0)
		{
			vector<AgentSearchPoint> prev;

			bool solnfound = pathfind_naive4(pos + possibilities[i].path, player, obstacles, visForces, force, prev, possibilities[i].g);

			if(solnfound)
			{
				force = possibilities[i].path;

				pair<Vector3, Vector3> vis(pos, pos + force);
				visForces.push_back(vis);

				return true;
			}
		}
	}

	force = Vector3(0,0,0);
	return false;
}

void CAgent::think(float delta, const AgentSearchStrategy strat, const vector<CAgent *> &agents, const CAgent &player, const vector<CAgent *> &obstacles, vector< pair <Vector3, Vector3> > &visForces, bool applyforce)
{
	_thinktimer += delta;
	/*
	// only think every once in a while
	if(_thinktimer > 0.100f)
		_thinktimer = 0.0f;
	else
	{
		_obj->actor->addForce(toNx(_force));
		return;
	}
	*/

	_force = Vector3::ZERO;
	Vector3 vel = toOgre(_obj->actor->getLinearVelocity());
	Vector3 dir = vel.normalisedCopy();

	// keep away from fellow agents - pedestrian forces model
	for(unsigned int i = 0; i < agents.size(); i++)
	{
		Vector3 fellowforce = attract(agents[i], 5.0, -100.0);
		_force += fellowforce;

		pair<Vector3, Vector3> vis(getPosition(), getPosition() + fellowforce);
		visForces.push_back(vis);
	}
	
	

	// keep away from obstacles - work in progress
	/*
	for(unsigned int i = 0; i < obstacles.size(); i++)
	{

		Ray ray(getPosition(), vel);
		pair< bool, Real > intersection = ray.intersects(aab);

		if(intersection.first && intersection.second < 5.0)
		{
			Vector3 wallforce = attract(obstacles[i], 20.0, -500.0);
			force += wallforce;

			pair<Vector3, Vector3> vis(getPosition(), getPosition() + wallforce);
			visForces.push_back(vis);
		}
	}
	*/

	Vector3 playerforce;

	vector<AgentSearchPoint> prev;

	if(strat == AgentSearchStrategyNaiveGrid)
		pathfind_naive(getPosition(), player, obstacles, visForces, playerforce);
	else if(strat == AgentSearchStrategyNaiveGrid4)
		pathfind_naive4(getPosition(), player, obstacles, visForces, playerforce, prev);
	else if(strat == AgentSearchStrategyRays)
		pathfind_ray(getPosition(), player, obstacles, visForces, playerforce);
	else if(strat == AgentSearchStrategyRayCombo)
		pathfind_raycombo(getPosition(), player, obstacles, visForces, playerforce);
	
	playerforce = playerforce.normalisedCopy();

	// rotate playerforce so it corrects for our current velocity?

	_force += playerforce * 1000.0;
	

	// never apply a force in the y direction
	_force.y = 0.0;

	if(applyforce)
		_obj->actor->addForce(toNx(_force));
}





